# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import datetime
from django.utils.timezone import utc


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0008_bill_ocr_full_text'),
    ]

    operations = [
        migrations.AddField(
            model_name='bill',
            name='updated_at',
            field=models.DateTimeField(default=datetime.datetime(
                2016, 1, 20, 18, 38, 27, 47441, tzinfo=utc), auto_now=True),
            preserve_default=False,
        ),
    ]



from councilmatic_core.models import Bill
from haystack import indexes
from councilmatic_core.templatetags.extras import clean_html

# XXX: is it OK to link to Django settings in haystack_indexes.py ?
from django.conf import settings
import pytz
app_timezone = pytz.timezone(settings.TIME_ZONE)


class BillIndex(indexes.SearchIndex):

    text = indexes.CharField(document=True, use_template=True,
                             template_name="search/indexes/councilmatic_core/bill_text.txt")
    slug = indexes.CharField(model_attr='slug', indexed=False)
    ocd_id = indexes.CharField(model_attr='ocd_id', indexed=False)
    bill_type = indexes.CharField(faceted=True)
    identifier = indexes.CharField(model_attr='identifier')
    description = indexes.CharField(model_attr='description', boost=1.25)
    source_url = indexes.CharField(model_attr='source_url', indexed=False)
    source_note = indexes.CharField(model_attr='source_note')
    abstract = indexes.CharField(model_attr='abstract', boost=1.25, default='')

    friendly_name = indexes.CharField()
    sort_name = indexes.CharField()
    sponsorships = indexes.MultiValueField(faceted=True)
    actions = indexes.MultiValueField()
    controlling_body = indexes.MultiValueField(faceted=True)
    full_text = indexes.CharField(model_attr='full_text', default='')
    ocr_full_text = indexes.CharField(model_attr='ocr_full_text', default='')
    last_action_date = indexes.DateTimeField()
    inferred_status = indexes.CharField(faceted=True)
    legislative_session = indexes.CharField(faceted=True)

    def get_model(self):
        return Bill

    def prepare_friendly_name(self, obj):
        return obj.friendly_name

    def prepare_sort_name(self, obj):
        return obj.friendly_name.replace(" ", "")

    def prepare_bill_type(self, obj):
        return obj.bill_type.lower()

    def prepare_sponsorships(self, obj):
        return [sponsorship.person for sponsorship in obj.sponsorships.all()]

    def prepare_actions(self, obj):
        return [action for action in obj.actions.all()]

    def prepare_controlling_body(self, obj):
        if obj.controlling_body:
            return [org.name for org in obj.controlling_body]

    def prepare_full_text(self, obj):
        return clean_html(obj.full_text)

    def prepare_last_action_date(self, obj):
        from datetime import datetime, timedelta
        if not obj.last_action_date:
            return datetime.now().replace(tzinfo=app_timezone) - timedelta(days=36500)
        return obj.last_action_date

    def prepare_inferred_status(self, obj):
        return obj.inferred_status

    def prepare_legislative_session(self, obj):
        return obj._legislative_session.identifier

    def prepare_ocr_full_text(self, obj):
        return clean_html(obj.ocr_full_text)

    def get_updated_field(self):
        return 'updated_at'



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-10-19 17:24
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0011_auto_20161013_1541'),
        ('councilmatic_core', '0026_auto_20161017_1641'),
    ]

    operations = [
    ]



from datetime import datetime
import inspect
import importlib

import pytz

from django.db import models
from django.core.exceptions import ImproperlyConfigured
from django.conf import settings
from django.core.urlresolvers import reverse
from django.core.urlresolvers import NoReverseMatch
from django.utils import timezone

if not (hasattr(settings, 'OCD_CITY_COUNCIL_ID') or hasattr(settings, 'OCD_CITY_COUNCIL_NAME')):
    raise ImproperlyConfigured(
        'You must define a OCD_CITY_COUNCIL_ID or OCD_CITY_COUNCIL_NAME in settings.py')

if not hasattr(settings, 'CITY_COUNCIL_NAME'):
    raise ImproperlyConfigured(
        'You must define a CITY_COUNCIL_NAME in settings.py')

MANUAL_HEADSHOTS = settings.MANUAL_HEADSHOTS if hasattr(
    settings, 'MANUAL_HEADSHOTS') else {}

app_timezone = pytz.timezone(settings.TIME_ZONE)

bill_document_choices = (
    ('A', 'Attachment'),
    ('V', 'Version'),
)


def override_relation(base_model):

    models_module = '{0}.models'.format(settings.APP_NAME)
    app_models = importlib.import_module(models_module)

    overridden = base_model
    for name, member in inspect.getmembers(app_models):

        try:
            if issubclass(member, base_model.__class__):
                module = __import__(models_module, globals(), locals(), [name])

                overridden = getattr(module, name)()

                attrs = [
                    f.attname for f in base_model.__class__._meta.fields] + ['_state']
                for attr in attrs:
                    setattr(overridden, attr, getattr(base_model, attr))

        except TypeError:
            pass

    return overridden


def get_uuid():
    import uuid
    return str(uuid.uuid4())


class Person(models.Model):
    ocd_id = models.CharField(max_length=100, unique=True, default=get_uuid, primary_key=True)
    name = models.CharField(max_length=100)
    headshot = models.CharField(max_length=255, blank=True, null=True)
    source_url = models.CharField(max_length=255, blank=True, null=True)
    source_note = models.CharField(max_length=255, blank=True, null=True)
    website_url = models.CharField(max_length=255, blank=True, null=True)
    email = models.CharField(max_length=255, blank=True, null=True)
    slug = models.CharField(max_length=255, unique=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.name

    @property
    def latest_council_membership(self):
        if hasattr(settings, 'OCD_CITY_COUNCIL_ID'):
            filter_kwarg = {'_organization__ocd_id': settings.OCD_CITY_COUNCIL_ID}
        else:
            filter_kwarg = {'_organization__name': settings.OCD_CITY_COUNCIL_NAME}

        city_council_memberships = self.memberships.filter(**filter_kwarg)

        if city_council_memberships.count():
            return city_council_memberships.order_by('-start_date', '-end_date').first()

        return None

    @property
    def current_council_seat(self):
        m = self.latest_council_membership
        if m and m.post:
            if self == m.post.current_member.person:
                if m.post:
                    return m.post.label
        return ''

    @property
    def latest_council_seat(self):
        m = self.latest_council_membership
        if m and m.post:
            return m.post.label
        return ''

    @property
    def is_speaker(self):
        return True if self.memberships.filter(role='Speaker').first() else False

    @property
    def headshot_url(self):
        if self.slug in MANUAL_HEADSHOTS:
            return '/static/images/' + MANUAL_HEADSHOTS[self.slug]['image']
        elif self.headshot:
            return '/static/images/' + self.ocd_id + ".jpg"
        else:
            return '/static/images/headshot_placeholder.png'

    @property
    def headshot_source(self):
        if self.slug in MANUAL_HEADSHOTS:
            return MANUAL_HEADSHOTS[self.slug]['source']
        elif self.headshot:
            return settings.CITY_VOCAB['SOURCE']
        else:
            return None

    @property
    def link_html(self):

        if self.ocd_id and self.slug:

            try:
                link_path = reverse('{}:person'.format(settings.APP_NAME), args=(self.slug,))

            except NoReverseMatch:
                link_path = reverse('person', args=(self.slug,))

            return '<a href="{0}" title="More on {1}">{1}</a>'.format(link_path, self.name)

        return self.name

    @property
    def primary_sponsorships(self):
        return self.sponsorships.filter(is_primary=True)

    @property
    def chair_role_memberships(self):
        if hasattr(settings, 'COMMITTEE_CHAIR_TITLE'):
            return self.memberships.filter(role=settings.COMMITTEE_CHAIR_TITLE)
        else:
            return []

    @property
    def member_role_memberships(self):
        if hasattr(settings, 'COMMITTEE_MEMBER_TITLE'):
            return self.memberships.filter(role=settings.COMMITTEE_MEMBER_TITLE)
        else:
            return []


class Bill(models.Model):
    ocd_id = models.CharField(max_length=100, unique=True, primary_key=True)
    ocd_created_at = models.DateTimeField(default=None)
    ocd_updated_at = models.DateTimeField(default=None)
    description = models.TextField()
    identifier = models.CharField(max_length=50)
    bill_type = models.CharField(max_length=50)
    classification = models.CharField(max_length=100)
    source_url = models.CharField(max_length=255)
    source_note = models.CharField(max_length=255, blank=True)
    subject = models.CharField(max_length=255, blank=True, null=True)

    _from_organization = models.ForeignKey('Organization',
                                           related_name='bills',
                                           null=True,
                                           db_column='from_organization_id')

    full_text = models.TextField(blank=True, null=True)
    ocr_full_text = models.TextField(blank=True, null=True)
    abstract = models.TextField(blank=True, null=True)
    last_action_date = models.DateTimeField(default=None, null=True)
    updated_at = models.DateTimeField(auto_now=True)

    _legislative_session = models.ForeignKey('LegislativeSession',
                                             related_name='bills',
                                             null=True,
                                             db_column='legislative_session_id')

    slug = models.CharField(max_length=255, unique=True)

    @property
    def from_organization(self):
        return override_relation(self._from_organization)

    @property
    def legislative_session(self):
        return override_relation(self._legislative_session)

    def __str__(self):
        return self.friendly_name

    @property
    def controlling_body(self):
        """
        grabs the organization that's currently 'responsible' for a bill
        """

        if self.current_action:
            related_orgs = self.current_action.related_entities.filter(
                entity_type='organization').all()
            # when a bill is referred from city council
            # to a committee, controlling body is the organization
            # the bill was referred to (a related org)
            if related_orgs:
                controlling_bodies = [Organization.objects.get(
                    ocd_id=org.organization_ocd_id) for org in related_orgs]
                return controlling_bodies
            # otherwise, the controlling body is usually whatever organization
            # performed the most recent action (this is the case most of the
            # time)
            else:
                return [self.current_action.organization]
        else:
            return None

    @property
    def last_action_org(self):
        """
        grabs whatever organization performed the most recent action
        """
        return self.current_action.organization if self.current_action else None

    @property
    def ordered_actions(self):
        """
        returns all actions ordered by date in descending order
        """
        return self.actions.all().order_by('-order')

    @property
    def current_action(self):
        """
        grabs the most recent action on a bill
        """
        return self.actions.all().order_by('-order').first() if self.actions.all() else None

    @property
    def first_action(self):
        """
        grabs the first action on a bill
        """
        return self.actions.all().order_by('order').first() if self.actions.all() else None

    @property
    def date_passed(self):
        return self.actions.filter(classification='passage').order_by('-order').first().date if self.actions.all() else None

    @property
    def friendly_name(self):
        """
        the bill title/headers displayed throughout the site (bill listings, bill detail pages)

        by default this returns the bill identifier - override this in
        custom subclass to construct a friendly name that makes sense locally
        """
        return self.identifier

    @property
    def primary_sponsor(self):
        """
        grabs the primary sponsorship for a bill
        """
        return self.sponsorships.filter(is_primary=True).first()

    @property
    def pseudo_topics(self):
        """
        returns a list of artificial topics for a bill, from the committees
        that have been involved in the bill's history (the actions)

        this serves as a backup when there isn't data on the real topics,
        so that bill listings can still have some useful tags populated
        """
        if self.actions.all():

            orgs = set([a.organization.name for a in self.actions.all() if
                        (a.organization.name != 'Mayor' and
                         a.organization.name != settings.CITY_COUNCIL_NAME)])

            if not orgs and self.controlling_body and \
                    self.controlling_body[0].name != settings.CITY_COUNCIL_NAME:

                orgs = self.controlling_body

            return list(orgs)
        else:
            return []

    @property
    def topics(self):
        """
        returns a list of topics for a bill

        override this in custom subclass for richer topic logic
        """
        return []

    @property
    def addresses(self):
        """
        returns a list of relevant addresses for a bill

        override this in custom subclass
        """
        return []

    @property
    def inferred_status(self):
        """
        infers bill status, to be displayed in colored labels next to bill names

        override this in custom subclass w/ richer logic for determining
        bill status, e.g. active, passed, approved, failed, stale
        """
        return None

    @property
    def listing_description(self):
        if self.abstract:
            return self.abstract
        return self.description

    @property
    def full_text_doc_url(self):
        """
        override this if instead of having full text as string stored in
        full_text, it is a PDF document that you can embed on the page
        """
        return None

    @property
    def attachments(self):
        """
        grabs that documents that are attachments (as opposed to versions)
        """
        return self.documents.filter(document_type='A').all()

    def get_last_action_date(self):
        """
        grabs date of most recent activity on a bill
        """
        return self.actions.all().order_by('-order').first().date if self.actions.all() else None

    @classmethod
    def bills_since(cls, date_cutoff):
        """
        grabs all bills that have had activity since a given date
        """
        return cls.objects.filter(last_action_date__gte=date_cutoff)

    @classmethod
    def new_bills_since(cls, date_cutoff):
        """
        grabs all bills that have been added since a given date
        (bills_since = new_bills_since + updated_bills_since)
        """
        all_bills_since = cls.bills_since(date_cutoff)
        new_bills_since = [
            b for b in all_bills_since if b.first_action.date >= date_cutoff]
        return new_bills_since

    @classmethod
    def updated_bills_since(cls, date_cutoff):
        """
        grabs all previously existing bills that have had activity since a given date
        (bills_since = new_bills_since + updated_bills_since)
        """
        all_bills_since = cls.bills_since(date_cutoff)
        updated_bills_since = [
            b for b in all_bills_since if b.first_action.date < date_cutoff]
        return updated_bills_since

    @property
    def unique_related_upcoming_events(self):
        events = [r.agenda_item.event for r in self.related_agenda_items.filter(
            agenda_item__event__start_time__gte=timezone.now(app_timezone)).all()]
        return list(set(events))


class Organization(models.Model):
    ocd_id = models.CharField(max_length=100, unique=True, primary_key=True)
    name = models.CharField(max_length=255)
    classification = models.CharField(max_length=255, null=True)
    _parent = models.ForeignKey(
        'self', related_name='children', null=True, db_column='parent_id')
    source_url = models.CharField(max_length=255, blank=True, null=True)
    slug = models.CharField(max_length=255, unique=True)
    updated_at = models.DateTimeField(auto_now=True)

    @property
    def parent(self):
        return override_relation(self._parent)

    def __str__(self):
        return self.name

    @classmethod
    def committees(cls):
        """
        grabs all organizations (1) classified as a committee & (2) with at least one member
        """
        return cls.objects.filter(classification='committee').order_by('name').filter(memberships__isnull=False).distinct()

    @property
    def recent_activity(self):
        # setting arbitrary max of 300 b/c otherwise page will take forever to
        # load
        return self.actions.order_by('-date', '-_bill__identifier', '-order')[:300]

    @property
    def recent_events(self):
        # need to look up event participants by name
        events = Event.objects.filter(participants__entity_type='organization', participants__entity_name=self.name)
        events = events.order_by('-start_time').all()
        return events

    @property
    def upcoming_events(self):
        """
        grabs events in the future
        """
        # need to look up event participants by name
        events = Event.objects\
                      .filter(participants__entity_type='organization', participants__entity_name=self.name)\
                      .filter(start_time__gt=datetime.now(app_timezone))\
                      .order_by('start_time')\
                      .all()
        return events

    @property
    def chairs(self):
        if hasattr(settings, 'COMMITTEE_CHAIR_TITLE'):
            return self.memberships.filter(role=settings.COMMITTEE_CHAIR_TITLE)
        else:
            return []

    @property
    def non_chair_members(self):
        if hasattr(settings, 'COMMITTEE_MEMBER_TITLE'):
            return self.memberships.filter(role=settings.COMMITTEE_MEMBER_TITLE)
        else:
            return []

    @property
    def link_html(self):
        link_fmt = '<a href="{0}">{1}</a>'

        if self.classification == 'committee':

            try:
                link_path = reverse('{}:committee_detail'.format(settings.APP_NAME), args=(self.slug,))
            except NoReverseMatch:
                link_path = reverse('committee_detail', args=(self.slug,))

            return link_fmt.format(link_path, self.name)

        if self.classification == 'legislature':

            try:
                link_path = reverse('{}:council_members'.format(settings.APP_NAME))
            except NoReverseMatch:
                link_path = reverse('council_members')

            return link_fmt.format(link_path, self.name)

        return self.name


class Action(models.Model):
    date = models.DateTimeField(default=None)
    classification = models.CharField(max_length=100, null=True)
    description = models.TextField(blank=True)

    _organization = models.ForeignKey('Organization',
                                      related_name='actions',
                                      null=True,
                                      db_column='organization_id')

    _bill = models.ForeignKey('Bill',
                              related_name='actions',
                              null=True,
                              db_column='bill_id')

    order = models.IntegerField()
    updated_at = models.DateTimeField(auto_now=True)

    @property
    def bill(self):
        return override_relation(self._bill)

    @property
    def organization(self):
        return override_relation(self._organization)

    @property
    def related_organization(self):
        r = self.related_entities.first()
        if r and r.entity_type == 'organization':
            org = Organization.objects.filter(
                ocd_id=r.organization_ocd_id).first()
            return org
        else:
            return None

    @property
    def label(self):
        c = self.classification

        if c == 'committee-passage':
            return 'success'
        if c == 'passage':
            return 'success'
        if c == 'executive-signature':
            return 'success'
        if c == 'amendment-passage':
            return 'success'

        if c == 'amendment-introduction':
            return 'default'
        if c == 'introduction':
            return 'default'
        if c == 'committee-referral':
            return 'default'
        if c == 'filing':
            return 'default'
        if c == 'executive-received':
            return 'default'

        if c == 'deferred':
            return 'primary'

        else:
            return 'default'

    @classmethod
    def actions_on_date(cls, date_match):
        """
        grabs all actions that occurred on a day
        """
        return cls.objects.filter(date__startswith=date_match)


class ActionRelatedEntity(models.Model):
    _action = models.ForeignKey(
        'Action', related_name='related_entities', db_column='action_id', null=True)
    entity_type = models.CharField(max_length=100)
    entity_name = models.CharField(max_length=255)
    organization_ocd_id = models.CharField(max_length=100, blank=True, null=True)
    person_ocd_id = models.CharField(max_length=100, blank=True, null=True)
    updated_at = models.DateTimeField(auto_now=True)

    @property
    def action(self):
        return override_relation(self._action)


class Post(models.Model):
    ocd_id = models.CharField(max_length=100, unique=True, primary_key=True)
    label = models.CharField(max_length=255)
    role = models.CharField(max_length=255)
    _organization = models.ForeignKey(
        'Organization', related_name='posts', db_column='organization_id', null=True)
    shape = models.TextField(blank=True, null=True)
    division_ocd_id = models.CharField(max_length=255, null=True)
    updated_at = models.DateTimeField(auto_now=True)

    @property
    def organization(self):
        return override_relation(self._organization)

    @property
    def current_member(self):
        if self.memberships.all():
            most_recent_member = self.memberships.order_by(
                '-end_date', '-start_date').first()
            if most_recent_member.end_date:
                return None
            else:
                return most_recent_member
        else:
            return None


class Membership(models.Model):
    _organization = models.ForeignKey(
        'Organization', related_name='memberships', db_column='organization_id', null=True)
    _person = models.ForeignKey(
        'Person', related_name='memberships', db_column='person_id', null=True)
    _post = models.ForeignKey(
        'Post', related_name='memberships', null=True, db_column='post_id')
    label = models.CharField(max_length=255, blank=True)
    role = models.CharField(max_length=255, blank=True)
    start_date = models.DateField(default=None, null=True)
    end_date = models.DateField(default=None, null=True)
    updated_at = models.DateTimeField(auto_now=True)

    @property
    def organization(self):
        return override_relation(self._organization)

    @property
    def person(self):
        return override_relation(self._person)

    @property
    def post(self):
        return override_relation(self._post)


class Sponsorship(models.Model):
    _bill = models.ForeignKey(
        'Bill', related_name='sponsorships', db_column='bill_id', null=True)
    _person = models.ForeignKey(
        'Person', related_name='sponsorships', db_column='person_id', null=True)
    classification = models.CharField(max_length=255)
    is_primary = models.BooleanField(default=False)
    updated_at = models.DateTimeField(auto_now=True)

    @property
    def bill(self):
        return override_relation(self._bill)

    @property
    def person(self):
        return override_relation(self._person)

    def __str__(self):
        return '{0} ({1})'.format(self.bill.identifier, self.person.name)


class Event(models.Model):
    ocd_id = models.CharField(max_length=100, unique=True, primary_key=True)
    ocd_created_at = models.DateTimeField(default=None)
    ocd_updated_at = models.DateTimeField(default=None)
    name = models.CharField(max_length=255)
    description = models.TextField()
    classification = models.CharField(max_length=100)
    start_time = models.DateTimeField()
    end_time = models.DateTimeField(null=True)
    all_day = models.BooleanField(default=False)
    status = models.CharField(max_length=100)
    location_name = models.CharField(max_length=255)
    location_url = models.CharField(max_length=255, blank=True)
    source_url = models.CharField(max_length=255)
    source_note = models.CharField(max_length=255, blank=True)
    slug = models.CharField(max_length=255, unique=True)
    updated_at = models.DateTimeField(auto_now=True)

    @property
    def event_page_url(self):

        try:
            link = reverse('{}:event_detail'.format(settings.APP_NAME), args=(self.slug,))
        except NoReverseMatch:
            link = reverse('event_detail', args=(self.slug,))

        return link

    @property
    def link_html(self):
        return '<a href="{0}" title="View Event Details">{1}</a>'.format(self.event_page_url, self.name)

    @property
    def clean_agenda_items(self):
        agenda_items = self.agenda_items.order_by('order').all()
        agenda_deduped = []
        descriptions_seen = []
        for a in agenda_items:
            if a.description not in descriptions_seen:
                descriptions_seen.append(a.description)
                agenda_deduped.append(a)

        return agenda_deduped

    @classmethod
    def next_city_council_meeting(cls):
        if hasattr(settings, 'CITY_COUNCIL_MEETING_NAME'):
            return cls.objects.filter(name__icontains=settings.CITY_COUNCIL_MEETING_NAME)\
                .filter(start_time__gt=timezone.now()).order_by('start_time').first()
        else:
            return None

    @classmethod
    def most_recent_past_city_council_meeting(cls):
        if hasattr(settings, 'CITY_COUNCIL_MEETING_NAME'):
            return cls.objects.filter(name__icontains=settings.CITY_COUNCIL_MEETING_NAME)\
                .filter(start_time__lt=timezone.now()).order_by('-start_time').first()
        else:
            return None

    @classmethod
    def upcoming_committee_meetings(cls):
        return cls.objects.filter(start_time__gt=timezone.now())\
                  .exclude(name__icontains=settings.CITY_COUNCIL_MEETING_NAME)\
                  .order_by('start_time').all()[:3]


class EventParticipant(models.Model):
    event = models.ForeignKey('Event', related_name='participants')
    note = models.TextField()
    entity_name = models.CharField(max_length=255)
    entity_type = models.CharField(max_length=100)
    updated_at = models.DateTimeField(auto_now=True)

    def __init__(self, *args, **kwargs):
        super(EventParticipant, self).__init__(*args, **kwargs)
        self.event = override_relation(self.event)


class EventAgendaItem(models.Model):
    event = models.ForeignKey('Event', related_name='agenda_items')
    order = models.IntegerField()
    description = models.TextField()
    bill = models.ForeignKey('Bill', related_name='related_agenda_items', null=True)
    note = models.CharField(max_length=255, null=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __init__(self, *args, **kwargs):
        super(EventAgendaItem, self).__init__(*args, **kwargs)
        self.event = override_relation(self.event)
        self.bill = override_relation(self.bill)


class Document(models.Model):
    note = models.TextField()
    url = models.TextField(blank=True)
    full_text = models.TextField(blank=True, null=True)

    class Meta:
        abstract = True


class BillDocument(Document):
    bill = models.ForeignKey('Bill', related_name='documents')
    document_type = models.CharField(
        max_length=255, choices=bill_document_choices)
    updated_at = models.DateTimeField(auto_now=True)

    def __init__(self, *args, **kwargs):
        super(BillDocument, self).__init__(*args, **kwargs)
        self.bill = override_relation(self.bill)


class EventDocument(Document):
    event = models.ForeignKey('Event', related_name='documents')

    def __init__(self, *args, **kwargs):
        super(EventDocument, self).__init__(*args, **kwargs)
        self.event = override_relation(self.event)


class LegislativeSession(models.Model):
    identifier = models.CharField(max_length=255, primary_key=True)
    jurisdiction_ocd_id = models.CharField(max_length=255)
    name = models.CharField(max_length=255)
    updated_at = models.DateTimeField(auto_now=True)



from django.test import TestCase, Client

from councilmatic_core.models import Organization, Bill, Person, Event


class RouteTest(TestCase):
    fixtures = [
        'organization.json',
        'membership.json',
        'bill.json',
        'person.json',
        'post.json',
        'event.json',
    ]

    def getPage(self, url):
        client = Client()
        return client.get(url)

    def test_index(self):
        assert self.getPage('/').status_code == 200

    def test_about(self):
        assert self.getPage('/about/').status_code == 200

    def test_committees(self):
        assert self.getPage('/committees/').status_code == 200

    def test_council_members(self):
        assert self.getPage('/council-members/').status_code == 200

    def test_committee(self):
        for committee in Organization.committees():
            committee_url = '/committee/{}/'.format(committee.slug)
            assert self.getPage(committee_url).status_code == 200

    def test_committee_event_rss(self):
        for committee in Organization.committees():
            committee_url = '/committee/{}/events/rss/'.format(committee.slug)
            assert self.getPage(committee_url).status_code == 200

    def test_committee_action_rss(self):
        for committee in Organization.committees():
            committee_url = '/committee/{}/actions/rss/'.format(committee.slug)
            assert self.getPage(committee_url).status_code == 200

    def test_committee_widget(self):
        for committee in Organization.committees():
            committee_url = '/committee/{}/widget/'.format(committee.slug)
            assert self.getPage(committee_url).status_code == 200

    def test_bill(self):
        for bill in Bill.objects.all():
            url = '/legislation/{}/'.format(bill.slug)
            assert self.getPage(url).status_code == 200

    def test_bill_rss(self):
        for bill in Bill.objects.all():
            url = '/legislation/{}/rss/'.format(bill.slug)
            assert self.getPage(url).status_code == 200

    def test_bill_widget(self):
        for bill in Bill.objects.all():
            url = '/legislation/{}/widget/'.format(bill.slug)
            assert self.getPage(url).status_code == 200

    def test_person(self):
        for person in Person.objects.all():
            url = '/person/{}/'.format(person.slug)
            assert self.getPage(url).status_code == 200

    def test_person_rss(self):
        for person in Person.objects.all():
            url = '/person/{}/rss/'.format(person.slug)
            assert self.getPage(url).status_code == 200

    def test_person_widget(self):
        for person in Person.objects.all():
            url = '/person/{}/widget/'.format(person.slug)
            assert self.getPage(url).status_code == 200

    def test_events(self):
        assert self.getPage('/events/').status_code == 200

    def test_event(self):
        for event in Event.objects.all():
            url = '/event/{}/'.format(event.slug)
            assert self.getPage(url).status_code == 200



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-10-17 17:01
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0023_auto_20161014_1648'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='agendaitembill',
            name='agenda_item',
        ),
        migrations.RemoveField(
            model_name='agendaitembill',
            name='bill',
        ),
        migrations.AddField(
            model_name='eventagendaitem',
            name='bill',
            field=models.ForeignKey(default='', on_delete=django.db.models.deletion.CASCADE, related_name='related_agenda_items', to='councilmatic_core.Bill'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='eventagendaitem',
            name='note',
            field=models.CharField(default='', max_length=255),
            preserve_default=False,
        ),
        migrations.DeleteModel(
            name='AgendaItemBill',
        ),
    ]



from django import template
from django.template.defaultfilters import stringfilter
from django.utils.html import strip_tags
import re

from django.utils.safestring import mark_safe
from django.core.serializers import serialize
import json
from django.db.models.query import QuerySet
import urllib

register = template.Library()


@register.filter
@stringfilter
def sentence_case(value):
    return value.replace("_", " ").capitalize()


@register.filter
@stringfilter
def inferred_status_label(status):
    return "<span class='label label-" + status.lower() + "'>" + status + "</span>"


@register.filter
@stringfilter
def facet_name(value):
    if value == 'bill_type':
        return 'Legislation type'
    if value == 'sponsorships':
        return 'Sponsor'
    if value == 'controlling_body':
        return 'Controlling body'
    if value == 'inferred_status':
        return 'Legislation status'


@register.filter
@stringfilter
def remove_action_subj(bill_action_desc):
    # removes 'by X' from bill action descriptions & expands abbrevs
    # for more readable action labels
    clean_action = re.sub(r'\bComm\b', 'Committee', bill_action_desc)
    clean_action = re.sub(r'\bRecved\b', 'Received', clean_action)
    clean_action = re.sub(r'[,\s]*by\s[^\s]*', '', clean_action)

    # shorten the really long action descriptions for approval w/ modifications
    if 'approved with modifications' in clean_action.lower():
        clean_action = 'Approved with Modifications'

    return clean_action


@register.filter
@stringfilter
def short_blurb(text_blob):
    if len(text_blob) > 196:
        blurb = text_blob[:196]
        blurb = blurb[:blurb.rfind(' ')] + ' ...'
        return blurb
    else:
        return text_blob


@register.filter
@stringfilter
def short_title(text_blob):
    if len(text_blob) > 28:
        blurb = text_blob[:24]
        blurb = blurb[:blurb.rfind(' ')] + ' ...'
        return blurb
    else:
        return text_blob


@register.filter
@stringfilter
def strip_mailto(email):
    return re.sub('mailto:', '', email)


@register.filter
@stringfilter
def committee_topic_only(committee_name):
    clean = re.sub('Committee on ', '', committee_name)
    clean = re.sub('Subcommittee on ', '', clean)
    if 'Mental Health, Developmental Disability' in clean:
        clean = 'Mental Health & Disability'
    return clean


@register.filter
@stringfilter
def clean_html(text):
    value = strip_tags(text).replace('\n', '')
    return re.sub(r'&(?:\w+|#\d+);', '', value)


@register.filter
@stringfilter
def alternative_identifiers(id_original):
    id_1 = re.sub(" ", " 0", id_original)
    id_2 = re.sub(" ", "", id_original)
    id_3 = re.sub(" ", "", id_1)
    return id_original + ' ' + id_1 + ' ' + id_2 + ' ' + id_3


@register.filter
def get_item(dictionary, key):
    return dictionary.get(key)


@register.filter
def format_date_sort(s, fmt='%Y%m%d%H%M'):
    if s:
        return s.strftime(fmt)
    else:
        return '0'


@register.filter
def format_url_parameters(url):
    params = ["?&sort_by=date", "?&sort_by=title", "?&sort_by=relevance", "?&ascending=true", "?&descending=true", "&sort_by=date", "&sort_by=title", "&sort_by=relevance", "&ascending=true", "&descending=true", "sort_by=date", "sort_by=title", "sort_by=relevance", "ascending=true", "descending=true"]

    paramsDict = dict((re.escape(el), "") for el in params)

    pattern = re.compile("|".join(paramsDict.keys()))

    return pattern.sub(lambda m: paramsDict[re.escape(m.group(0))], url)

# TODO: Clean up for refactor of javascript.
# @register.simple_tag
# def query_transform(request, **kwargs):

#     data_dict = dict(request.GET.copy())
#     print(data_dict)
#     try:
#         selected_facet_names = data_dict['selected_facets']
#     except:
#         selected_facet_names = []

#     updated = request.GET.copy()
#     if selected_facet_names:
#         for k,v in kwargs.items():
#             selected_facet_names.append(v)

#     updated['selected_facets'] = selected_facet_names

#     return updated.urlencode()



import urllib

from haystack.query import SearchQuerySet

from django.contrib.syndication.views import Feed
from django.utils.feedgenerator import Rss201rev2Feed
from django.core.urlresolvers import reverse, reverse_lazy
from django.conf import settings

from .models import Person, Bill, Organization, Event


class CouncilmaticFacetedSearchFeed(Feed):
    title_template = 'feeds/search_item_title.html'
    description_template = 'feeds/search_item_description.html'
    bill_model = Bill

    all_results = None
    sqs = SearchQuerySet().facet('bill_type')\
                          .facet('sponsorships', sort='index')\
                          .facet('controlling_body')\
                          .facet('inferred_status')
    query = None

    def url_with_querystring(self, path, **kwargs):
        return path + '?' + urllib.parse.urlencode(kwargs)

    def get_object(self, request):
        self.queryDict = request.GET
        all_results = SearchQuerySet().all()
        facets = None

        if 'selected_facets' in request.GET:
            facets = request.GET.getlist('selected_facets')

        if 'q' in request.GET:
            self.query = request.GET['q']
            results = all_results.filter(content=self.query)

            if facets:
                for facet in facets:
                    (facet_name, facet_value) = facet.split(':')
                    facet_name = facet_name.rsplit('_exact')[0]
                    results = results.narrow('%s:%s' % (facet_name, facet_value))
        elif facets:
            for facet in facets:
                (facet_name, facet_value) = facet.split(':')
                facet_name = facet_name.rsplit('_exact')[0]
                results = all_results.narrow('%s:%s' % (facet_name, facet_value))

        return results.order_by('-last_action_date')

    def title(self, obj):
        if self.query:
            title = settings.SITE_META['site_name'] + ": Search for '" + self.query.capitalize() + "'"
            # XXX: create a nice title based on all search parameters
        else:
            title = settings.SITE_META['site_name'] + ": Filtered Search"

        return title

    def link(self, obj):
        # return the main non-RSS search URL somehow
        # XXX maybe "quargs" - evz
        # return reverse('councilmatic_search', args=(searchqueryset=self.sqs,))
        url = self.url_with_querystring(reverse('{}:councilmatic_search_feed'.format(settings.APP_NAME)), q=self.query)
        return url

    def item_link(self, bill):
        return reverse('bill_detail', args=(bill.slug,))

    def item_pubdate(self, bill):
        return bill.last_action_date

    def description(self, obj):
        return "Bills returned from search"

    def items(self, searchresults):
        l_items = list(searchresults)[:20]
        # turn these into bills. XXX: should override in subclasses, e.g. NYCCouncilmaticFacetedSearchFeed,
        # to access methods like inferred_status()
        pks = [i.pk for i in l_items]
        bills = self.bill_model.objects.filter(pk__in=pks).order_by('-last_action_date')
        return list(bills)


class PersonDetailFeed(Feed):
    """The PersonDetailFeed provides an RSS feed for a given committee member,
    returning the most recent 20 bills for which they are the primary sponsor;
    and for each bill, the list of sponsores and the action history.
    """

    title_template = 'feeds/person_detail_item_title.html'
    description_template = 'feeds/person_detail_item_description.html'
    feed_type = Rss201rev2Feed
    NUM_RECENT_BILLS = 20

    def get_object(self, request, slug):
        o = Person.objects.get(slug=slug)
        return o

    def title(self, obj):
        return settings.SITE_META['site_name'] + ': ' + settings.CITY_VOCAB['COUNCIL_MEMBER'] + " %s: Recently Sponsored Bills" % obj.name

    def link(self, obj):
        # return the Councilmatic URL for the person
        # XXX maybe put this in models.py:Person.get_absolute_url() instead (https://docs.djangoproject.com/en/1.9/ref/models/instances/ , https://docs.djangoproject.com/en/1.9/ref/contrib/syndication/)
        return reverse('person', args=(obj.slug,))

    # Return sponsored legislation a la https://nyc.councilmatic.org/person/margaret-s-chin/
    def item_link(self, bill):
        # return the Councilmatic URL for the bill
        return reverse('bill_detail', args=(bill.slug,))

    def item_pubdate(self, bill):
        return bill.last_action_date

    def description(self, obj):
        return "Recent sponsored bills from " + obj.name + "."

    def items(self, person):
        sponsored_bills = [s.bill for s in person.primary_sponsorships.order_by('-_bill__last_action_date')[:10]]
        recent_sponsored_bills = sponsored_bills[:self.NUM_RECENT_BILLS]
        return recent_sponsored_bills


class CommitteeDetailEventsFeed(Feed):
    """The CommitteeDetailEventsFeed provides an RSS feed for a given committee,
    returning the most recent 20 events.
    """

    title_template = 'feeds/committee_events_item_title.html'
    description_template = 'feeds/committee_events_item_description.html'
    feed_type = Rss201rev2Feed
    NUM_RECENT_COMMITTEE_EVENTS = 20

    def get_object(self, request, slug):
        o = Organization.objects.get(slug=slug)
        return o

    def title(self, obj):
        return settings.SITE_META['site_name'] + ": " + obj.name + ": Recent Events"

    def link(self, obj):
        # return the Councilmatic URL for the committee
        return reverse('committee_detail', args=(obj.slug,))

    def item_link(self, event):
        # return the Councilmatic URL for the event
        return reverse('event_detail', args=(event.slug,))

    def item_pubdate(self, event):
        return event.start_time

    def description(self, obj):
        return "Events for committee %s" % obj.name

    def items(self, obj):
        events = obj.recent_events.all()[:self.NUM_RECENT_COMMITTEE_EVENTS]
        levents = list(events)
        return levents


class CommitteeDetailActionFeed(Feed):
    """The CommitteeDetailActionFeed provides an RSS feed for a given committee,
    returning the most recent 20 actions on legislation.
    """

    # instead of defining item_title() or item_description(), use templates
    title_template = 'feeds/committee_actions_item_title.html'
    description_template = 'feeds/committee_actions_item_description.html'
    feed_type = Rss201rev2Feed
    NUM_RECENT_COMMITTEE_ACTIONS = 20

    def get_object(self, request, slug):
        o = Organization.objects.get(slug=slug)
        return o

    def title(self, obj):
        return settings.SITE_META['site_name'] + ": " + obj.name + ": Recent Actions"

    def link(self, obj):
        # return the Councilmatic URL for the committee
        return reverse('committee_detail', args=(obj.slug,))

    def item_link(self, action):
        # return the Councilmatic URL for the bill
        return reverse('bill_detail', args=(action.bill.slug,))

    def item_pubdate(self, action):
        return action.date

    def description(self, obj):
        return "Actions for committee %s" % obj.name

    def items(self, obj):
        actions = obj.recent_activity[:self.NUM_RECENT_COMMITTEE_ACTIONS]
        actions_list = list(actions)
        return actions_list


class BillDetailActionFeed(Feed):
    """
    Return the last 20 actions for a given bill.
    """

    # instead of defining item_title() or item_description(), use templates
    title_template = 'feeds/bill_actions_item_title.html'
    description_template = 'feeds/bill_actions_item_description.html'
    feed_type = Rss201rev2Feed
    NUM_RECENT_BILL_ACTIONS = 20

    def get_object(self, request, slug):
        o = Bill.objects.get(slug=slug)
        return o

    def title(self, obj):
        return settings.SITE_META['site_name'] + ": " + obj.friendly_name + ": Recent Actions"

    def link(self, obj):
        # return the Councilmatic URL for the committee
        return reverse('bill_detail', args=(obj.slug,))

    def item_link(self, action):
        # Bill actions don't have their own pages, so just link to the Bill page (?)
        return reverse('bill_detail', args=(action.bill.slug,))

    def item_pubdate(self, action):
        return action.date

    def description(self, obj):
        return "Actions for bill %s" % obj.friendly_name

    def items(self, obj):
        actions = obj.ordered_actions[:self.NUM_RECENT_BILL_ACTIONS]
        actions_list = list(actions)
        return actions_list


class EventsFeed(Feed):
    """
    Return the last 20 announced events as per, e.g., https://nyc.councilmatic.org/events/
    """
    title_template = 'feeds/events_item_title.html'
    description_template = 'feeds/events_item_description.html'
    feed_type = Rss201rev2Feed
    NUM_RECENT_EVENTS = 20

    title = settings.CITY_COUNCIL_NAME + " " + "Recent Events"
    link = reverse_lazy('events')
    description = "Recently announced events."

    def item_link(self, event):
        # return the Councilmatic URL for the event
        return reverse('event_detail', args=(event.slug,))

    def item_pubdate(self, event):
        return event.start_time

    def description(self, obj):
        return "Events"

    def items(self, obj):
        events = Event.objects.all()[:self.NUM_RECENT_EVENTS]
        levents = list(events)
        return levents



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-10-17 21:40
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0024_auto_20161017_1201'),
    ]

    operations = [
        migrations.AlterField(
            model_name='eventagendaitem',
            name='bill',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='related_agenda_items', to='councilmatic_core.Bill'),
        ),
    ]
